Module: Oxygene

Defined in:
lib/oxygene.rb,
lib/oxygene/cid.rb,
lib/oxygene/base32.rb,
lib/oxygene/errors.rb,
lib/oxygene/version.rb,
lib/oxygene/car_repo.rb,
lib/oxygene/extensions.rb,
lib/oxygene/car_archive.rb,
lib/oxygene/car_section.rb,
sig/oxygene.rbs

Overview

Various data decoding related primitives for working with AT Protocol repositories and firehose events.

This gem was extracted from Skyfall and is mostly used there for the CBOR firehose decoding, but can also be used standalone for other ATProto data handling purposes, like decoding downloaded .car account repos.

The functionality currently includes: Base32 encoder/decoder, CAR archive/repo parser, CID wrapper, and CBOR decoding (through the cbor gem for now).

The gem intentionally only includes support for the parts of these standards that are in use in ATProto, i.e. doesn't handle all possible types of CIDs defined in the CID standard, only generates lowercase Base32, and so on.

Defined Under Namespace

Modules: Base32, Extensions Classes: CARArchive, CARRepo, CARSection, CID, DecodeError, Error, UnsupportedError

Constant Summary collapse

VERSION =

Returns:

  • (String)
'0.1.0'