Exception: Oxygene::UnsupportedError

Inherits:
Error
  • Object
show all
Defined in:
lib/oxygene/errors.rb

Overview

Raised when a piece of data is technically formatted correctly, but not in a version that this library currently supports (or that is used in ATProto).