Exception: Oxygene::UnsupportedError
- Defined in:
- lib/oxygene/errors.rb
Overview
Raised when a piece of data is technically formatted correctly, but not in a version that this library currently supports (or that is used in ATProto).
Raised when a piece of data is technically formatted correctly, but not in a version that this library currently supports (or that is used in ATProto).